#include <cstring>
#include <iostream>
#include <stdlib.h>
#include <time.h>

#include "sbsshcommon.h"
#include "sbsshkeystorage.h"
#include "sbsimplesftp.h"
#include "sbsftpcommon.h"
#include "sbutils.h"

using namespace SecureBlackbox;

#define OP_LIST		1
#define OP_DOWNLOAD	2
#define OP_UPLOAD	3

void SB_CALLBACK Client_OnKeyValidate(void * /* objData */, TObjectHandle /* hSender */, TElSSHKeyHandle hServerKey, int8_t * Validate)
{
	try
	{
		TElSSHKey ServerKey(hServerKey, false);
		TMessageDigest128 M128;
		ServerKey.get_FingerprintMD5(M128);
		std::string dgst;
		DigestToStr(M128, true, dgst);
		std::cout << "Server key received, fingerprint " << dgst << std::endl;
		*Validate = true; // NEVER do this. You MUST check the key validity somehow
	}
	catch (SBException E)
	{
		std::cout << "Unexpected error in OnKeyValidate event handler" << std::endl;
		std::cout << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
		*Validate = false;
	}
}

void Download(TElSimpleSFTPClient &Client, const std::string &remoteName, const std::string &localName)
{
	Client.DownloadFile(remoteName, localName, ftmOverwrite);
	std::cout << "Downloaded file " << remoteName << " to " << localName << std::endl;
}

void Upload(TElSimpleSFTPClient &Client, const std::string &localName, const std::string &remoteName)
{
	Client.UploadFile(localName, remoteName, ftmOverwrite);
	std::cout << "Uploaded file " << localName << " to " << remoteName << std::endl;

	// now adjust the attributes of the uploaded file
	TElSftpFileAttributes Attrs;
	time_t t = time(NULL);
	Attrs.set_CTime(t);
	Attrs.set_MTime(t);
	Attrs.set_ATime(t);
	Attrs.set_CATime(t);

	Attrs.set_IncludedAttributes(static_cast<TSBSftpAttributes>(f_saATime | f_saMTime | f_saCTime | f_saCATime));
	Client.SetAttributes(remoteName, Attrs);
}

void List(TElSimpleSFTPClient &Client, const std::string &remoteName)
{
	std::vector<uint8_t> Handle;
	Client.OpenDirectory(remoteName, Handle);
	TList Listings;
	Client.ReadDirectory(Handle, Listings);
	if (Listings.get_Count() > 0)
	{
		std::cout << "Contents of " << remoteName << ":" << std::endl;
		for (int i = 0; i < Listings.get_Count(); i++)
		{
			TElSftpFileInfo Info((TElClassHandle)Listings.get_Items(i), false); // casting to TElSftpFileInfo
			std::string name;
			Info.get_Name(name);
			std::cout << name << std::endl;
		}
	}
	else
		std::cout << "The specified directory is empty" << std::endl;

	Client.CloseHandle(Handle);
}

int main(int argc, char **argv)
{
	try
	{
		#error Please pick the evaluation license key from <SecureBlackbox>\LicenseKey.txt file and place it here. If the evaluation key expires, you can request an extension using the form on https://www.eldos.com/sbb/keyreq/
SetLicenseKey("<put the key here>");
	}
	catch (SBException E)
	{
		std::cout << "Failed to set license key:" << std::endl;
		std::cout << E.what() << std::endl;
	}

	if (argc < 3)
	{
		std::cout << "Usage:" << std::endl;
		std::cout << "ConsoleSFTP {-download|-upload|-list} -addr[ess] <address> [-port <port>] -user[name] <user> -pass[word] <password> -remote <remote file name> -local <local file name>" << std::endl;
		return 0;
	}

	int paramOperation = 0;
	int paramPort = 22;
	std::string paramAddress;
	std::string paramUsername;
	std::string paramPassword;
	std::string paramLocalName;
	std::string paramRemoteName;

	int i = 1;
	while (i < argc)
	{
		char *arg = argv[i];
		if ((!strcmp(arg, "-addr") ||
			!strcmp(arg, "-address")) && (i < argc - 1))
			paramAddress.assign(argv[++i]);
		else if (!strcmp(arg, "-port") && (i < argc - 1))
		{
			paramPort = atoi(argv[++i]);
			if (0 == paramPort)
				paramPort = 22;
		}
		else if ((!strcmp(arg, "-user") ||
			!strcmp(arg, "-username")) && (i < argc - 1))
			paramUsername.assign(argv[++i]);
		else if ((!strcmp(arg, "-pass") ||
			!strcmp(arg, "-password")) && (i < argc - 1))
			paramPassword.assign(argv[++i]);
		else if (!strcmp(arg, "-down") ||
			!strcmp(arg, "-download"))
			paramOperation = OP_DOWNLOAD;
		else if (!strcmp(arg, "-up") ||
			!strcmp(arg, "-upload"))
			paramOperation = OP_UPLOAD;
		else if (!strcmp(arg, "-ls") ||
			!strcmp(arg, "-list"))
			paramOperation = OP_LIST;
		else if (!strcmp(arg, "-local") && (i < argc - 1))
			paramLocalName.assign(argv[++i]);
		else if (!strcmp(arg, "-remote") && (i < argc - 1))
			paramRemoteName.assign(argv[++i]);

		i++;
	}

	TElSimpleSFTPClient Client(NULL);
	Client.set_Address(paramAddress);
	Client.set_Port(paramPort);

	Client.set_Username(paramUsername);
	Client.set_Password(paramPassword);

	Client.set_OnKeyValidate(&Client_OnKeyValidate, NULL);

	try
	{
		std::cout << "Connecting " << paramAddress << ":" << paramPort << std::endl;
		Client.Open();
	}
	catch (SBException E)
	{
		std::cout << "Connection failed due to exception:" << std::endl;
		std::cout << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;

		std::cout << "If you have ensured that all connection parameters are correct and you still can't connect," << std::endl;
		std::cout << "please contact EldoS support as described on http://www.eldos.com/sbb/support-tech.php" << std::endl;
		std::cout << "Remember to provide details about the error that happened." << std::endl;

		std::string sname = "";
		Client.get_ServerSoftwareName(sname);
		if (sname.length() > 0)
			std::cout << "Server software identified itself as: " << sname.c_str() << std::endl;

		try
		{
			Client.Close(true);
		}
		catch (...)
		{
		}

		return 0;
	}

	try
	{
		switch (paramOperation)
		{
		case OP_DOWNLOAD:
			if (paramRemoteName == "")
			{
				std::cout << "Remote filename is not specified" << std::endl;
				break;
			}

			if (paramLocalName == "")
				paramLocalName = paramRemoteName;

			Download(Client, paramRemoteName, paramLocalName);
			break;

		case OP_UPLOAD:
			if (paramLocalName == "")
			{
				std::cout << "Local filename is not specified" << std::endl;
				break;
			}

			if (paramRemoteName == "")
				paramRemoteName = paramLocalName;

			Upload(Client, paramLocalName, paramRemoteName);
			break;

		default:
			if (paramRemoteName == "")
				paramRemoteName = "./";

			List(Client, paramRemoteName);
			break;
		}
	}
	catch (SBException E)
	{
		std::cout << E.what() << std::endl;
	}

	try
	{
		Client.Close(false);
	}
	catch (SBException E)
	{
		std::cout << E.what() << std::endl;
	}

	return 0;
}
